#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdarg.h>

#include "ioutils.h"

// returns zero on success
int write_string(int fd, const char *fmt, ...)
{
  char buf[1000];
  va_list ap;
  va_start(ap, fmt);
  int len = vsnprintf(buf, 1000, fmt, ap);
  va_end(ap);

  //  printf("WRITE: %s", buf);
  return write_fully(fd, buf, len) != len;
}
 
int write_fully(int fd, const void *b, int len)
{
  int cnt=0;
  int thiscnt;
  unsigned char *bb=(unsigned char*) b;

  /*
  printf("WRITE: ");
  for (int i = 0; i < len; i++)
    printf("%c", ((char*) b)[i]);
  printf("\n");
  */

  while (cnt<len)
    {
      thiscnt=write(fd, &bb[cnt], len-cnt);
      if (thiscnt<0) {
        perror ("write");
	return -1;
      }
      cnt+=thiscnt;
    }

  return cnt;
}

int read_fully(int fd, void *b, int len)
{
  int cnt=0;
  int thiscnt;
  unsigned char *bb=(unsigned char*) b;

  while (cnt<len)
    {
      thiscnt=read(fd, &bb[cnt], len-cnt);
      if (thiscnt<0) {
        perror ("read_fully");
	return -1;
      }
      if (thiscnt == 0) {
        fprintf (stderr, "read_fully end of file\n");
        return -1;
      }
      printf ("%d\n", thiscnt);
      cnt+=thiscnt;
    }

  return cnt;
}

/* returns -1 on error, 0 on timeout, nchar on success. */
int read_timeout(int fd, void *buf, int maxlen, int msTimeout)
{
  struct pollfd pfd;
  int len;
  int res;

 again:

  pfd.fd=fd;
  pfd.events=POLLIN;

  res=poll(&pfd, 1, msTimeout);
  if (res<0) // error
    {
      //      perror("read_timeout: poll");
      goto again; //return -1;
    }
  if (res==0) // timeout
    {
      errno=ETIMEDOUT;
      return 0;
    }

  //  printf("%d",pfd.revents==POLLIN);
  //  fflush(NULL);
  
  len=read(fd, buf, maxlen);
  if (len<0)
    {
      perror("read");
      return -1;
    }
  if (len == 0) {
    fprintf (stderr, "end of file\n");
    return -1;
  }
  return len;
}

/* returns -1 on error, nchar on timeout or success. */
int read_fully_timeout(int fd, void *bufin, int len, int msTimeout)
{
  char *buf=(char*) bufin;
  int readsofar=0;
  int thisread=0;

  while (readsofar<len)
    {
      thisread=read_timeout(fd, &buf[readsofar], len-readsofar, msTimeout);
      if (thisread == 0)
	return readsofar;
      if (thisread < 0)
	return thisread;

      readsofar+=thisread;
    }

  return len;
}

// reads number of bytes available, <0 on error.
int read_available(int fd)
{
  long avail = 0;

  if(ioctl(fd, FIONREAD, &avail) == 0)
    return avail;
  else
    return -1;
}

// read (and discard) all available data.
void read_flush(int fd)
{
  int avail;
  char buf[1024];

  do
    {
      avail=read_available(fd);
      if (avail>1024)
        avail=1024;

      read(fd, buf, avail);
    } while (avail>0);
}

// returns length of string, -1 on error, 0 on timeout.
int read_line_timeout_min(int fd, void *buf_in, int maxlen, int msTimeout)
{
  int len=0;
  int thislen;
  char *buf=(char*) buf_in;

  while (len<maxlen)
    {
      thislen=read_timeout(fd, &buf[len], 1, msTimeout);

      if (thislen < 0)
	return thislen;
      if (thislen==0)
	return 0;

      if (buf[len]=='\n' || buf[len]=='\r')
        {
          buf[len]=0;
	  if (len == 0)
	    continue;

          return len;
        }

      len++;
    }

  buf[len]=0;
  return len;
}


// returns length of string, -1 on error, 0 on timeout.
int read_line_timeout(int fd, void *buf_in, int maxlen, int msTimeout)
{
  int len=0;
  int thislen;
  char *buf=(char*) buf_in;

  while (len<maxlen)
    {
      thislen=read_timeout(fd, &buf[len], 1, msTimeout);
      if (thislen < 0)
	return thislen;
      if (thislen==0)
	return 0;

      if (buf[len]=='\n' || buf[len]=='\r')
        {
          buf[len]=0;
          return len;
        }

      len++;
    }

  buf[len]=0;
  return len;
}

int fwrite64(FILE *f, int64_t v64)
{
  int32_t v = v64>>32;
  fwrite32(f, v);
  v = v64 & 0xffffffff;
  fwrite32(f, v);
  return 0;
}

int fwrite32(FILE *f, int32_t v)
{
  v = htonl(v);
  fwrite(&v, 4, 1, f);
  return 0;
}

int fread32(FILE *f, int32_t *v32)
{
  int32_t v;

  if (fread(&v, 4, 1, f) != 1)
    return -1;

  *v32 = ntohl(v);

  return 0;
}

int fread64(FILE *f, int64_t *v64)
{
  int32_t v1, v2;

  if (fread32(f, &v1))
    return -1;

  if (fread32(f, &v2))
    return -1;

  *v64 =     (((int64_t) v1)<<32) | (((int64_t) v2)&0xffffffff);

  return 0;
}

