#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <poll.h>
#include <termios.h>
#include <arpa/inet.h>

#include "serial.h"
#include "getopt.h"
#include "ioutils.h"
#include "srec.h"
#include "uuutil.h"

#ifndef STDIN
#define STDIN 0
#endif

#define MAXLEN 1000

// XXX with an FTDI USB chip, writes occur really fast (faster than
// the actual baud rate). We can timeout prematurely if the timeout is
// set to reasonable values, because we start the timeout clock before
// the ftdi has finished sending the data.
#define TIMEOUT_MS 5000

// all data read/written in chunksize units must be less than 20 *
// MAX_UUENCODE_SIZE. Must be smaller than the smallest flash sector.
#define CHUNKSIZE 512

// maximum number of bytes to write in a single uuencode line.
// (the LPC supports a maximum of 20 lines at a time)
// must be less than 45 in order to observe maximum LPC line length.
#define MAX_UUENCODE_SIZE 32

// how big should the progress bars be?
#define PROGRESS_WIDTH 42

//////////////////////////////////////////////////////////////////////////
typedef struct 
{
    getopt_t *gopt;
    int fd;
    char *partid;
    int bootcode_version_major, bootcode_version_minor;
    struct device *device;
} lpcprog_t;

//////////////////////////////////////////////////////////////////////////
// stupid utility functions

int min(int a, int b)
{
    return a < b ? a : b;
}

int64_t utime()
{
    struct timeval tv;
    gettimeofday (&tv, NULL);
    return (int64_t) tv.tv_sec * 1000000 + tv.tv_usec;
}

chunk_t *chunk_create(int chunksize)
{
    chunk_t *c = (chunk_t*) calloc(1, sizeof(chunk_t));
    c->data = (char*) calloc(chunksize, sizeof(char));
    c->datalen = chunksize;
    return c;
}

uint32_t checksum_update(uint32_t v, uint32_t w)
{
    return ((v<<7) ^ (v>>25)) + w;
}

uint32_t chunks_checksum(chunk_t *c)
{
    uint32_t chk = 0;

    for ( ; c != NULL ; c = c->next)
    {
        uint32_t *p = (uint32_t*) c->data;

        // XXX assumes chunksize is divisble by 4
        for (int i = 0; i < c->datalen/4; i++)
            chk = checksum_update(chk, p[i]);
    }

    return chk;
}

uint32_t chunks_minaddr(chunk_t *c)
{
    uint32_t minaddr = 0xffffffff;

    for ( ; c != NULL ; c = c->next)
    {
        if (c->addr < minaddr)
            minaddr = c->addr;
    }

    return minaddr;
}

uint32_t chunks_maxaddr(chunk_t *c)
{
    uint32_t maxaddr = 0x0;

    for ( ; c != NULL ; c = c->next)
    {
        uint32_t t = c->addr + c->datalen - 1;

        if (t > maxaddr)
            maxaddr = t;
    }

    return maxaddr;
}


void fail(const char *s)
{
    printf("Failure: %s\n", s);
    exit(-1);
}

void warn(const char *s)
{
    printf("Warn: %s\n", s);
}


//////////////////////////////////////////////////////////////////////////
// (called by sync)
int lpc_version(lpcprog_t *lpc)
{
    int res;
    char line[MAXLEN];

    ////////// get version //////////////
    if (write_string(lpc->fd, "J\r\n"))
        return -1;

    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (line[0]!='0')
        return -1;

    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);

    if (lpc->partid != NULL)
        free(lpc->partid);
    lpc->partid = strdup(line);

    ////////// get bootloader version //////////////
    if (write_string(lpc->fd, "K\r\n"))
        return -1;

    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (line[0]!='0')
        return -1;

    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    lpc->bootcode_version_minor = atoi(line);
    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    lpc->bootcode_version_major = atoi(line);

    return 0;
}

//////////////////////////////////////////////////////////////////////////
int lpc_unlock(lpcprog_t *lpc)
{
    int res;
    char line[MAXLEN];

    if (write_string(lpc->fd, "U 23130\r\n"))
        return -1;

    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (line[0]!='0')
        return -1;

    return 0;
}

//////////////////////////////////////////////////////////////////////////
struct device *find_device(lpcprog_t *lpc)
{
    char *part = getopt_get_string(lpc->gopt, "part");
    if (!strcmp(part, "auto"))
        part = NULL;

    for (int idx = 0; devices[idx].numsectors != 0; idx++)
    {
        if (part != NULL && !strcmp(part, devices[idx].description))
            return &devices[idx];

        if (!strcmp(lpc->partid, devices[idx].partid))
            return &devices[idx];
    }

    fail ("Unable to find a device description that matches this part. Try specifing --part.");
    return NULL;
}

void list_devices(lpcprog_t *lpc)
{
    printf("Known devices: \n");

    for (int idx = 0; devices[idx].numsectors != 0; idx++)
    {
        if (idx%8==0)
            printf("    ");

        printf("%s ", devices[idx].description);
        if (idx%8==7)
            printf("\n");
    }
    printf("\n");

}

//////////////////////////////////////////////////////////////////////////
int lpc_sync(lpcprog_t *lpc)
{
    ssize_t res;
    char    line[MAXLEN];
    int     tries = 0;

    printf("Synchronizing with LPC...");
    fflush(NULL);

    strcpy(line, "");

sync:
    tries = 0;

    // if the user doesn't have the DTR/RTS magic, we should
    // prompt them to reset the board manually at this point.
    serial_set_dtr(lpc->fd, 1); // reset the board, holding p2.10 low
    serial_set_rts(lpc->fd, 1);
    usleep(100000);
    serial_set_dtr(lpc->fd, 0); // release reset, keep p2.10 low
    usleep(100000);
    serial_set_rts(lpc->fd, 0); // release p2.10

    printf(".");
    fflush(NULL);

    read_flush(lpc->fd);

 sync_again:
    tries++;
    if (tries > 5)
      goto sync;

    res = write(lpc->fd, "?", 1);
    if (res != 1)
        perror("write ?");

    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (res < 0) {
        goto sync_again;
    }
    
    if (strcmp(line, "Synchronized"))
        goto sync;

    if (write_string(lpc->fd, "Synchronized\r\n"))
        goto sync;

    // read and discard echo'ed "Synchronized"
    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);

    // read "OK"
    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (res < 0) {
        printf("OK read timeout");
        goto sync;
    }
    
    if (strcmp(line, "OK")) {
        goto sync;
    }

    if (write_string(lpc->fd, "%d\r\n", getopt_get_int(lpc->gopt, "khz")))
        goto sync;

    // discard echo'ed string
    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);

    // "read "OK"
    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (res < 0) {
        printf("OK read timeout");
        goto sync;
    }

    if (strcmp(line, "OK")) {
        goto sync;
    }

    // disable echo
    if (write_string(lpc->fd, "A 0\r\n"))
        goto sync;

    // discard echo'd string
    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);

    // get CMD_SUCCESS
    res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (line[0]!='0')
        goto sync;

    lpc_version(lpc);

    lpc->device = find_device(lpc);

    printf("Found device (part id %s, version %d.%d).\nUsing configuration %s.\n", 
           lpc->partid, 
           lpc->bootcode_version_major, lpc->bootcode_version_minor,
           lpc->device->description);

    return 0;
}

//////////////////////////////////////////////////////////////////////////
int lpc_read(lpcprog_t *lpc, uint32_t addr, int32_t len, char *buf)
{
    char line[MAXLEN];

    if (write_string(lpc->fd, "R %u %u\r\n", addr, len))
        return -1;

    read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (line[0]!='0') {
        printf("lpc_read: %s\n", line);
        return -1;
    }

    int readcount = 0;
    int thisread;
    int checksum=0;

receive:
    readcount = 0;
    checksum = 0;

    while (readcount < len)
    {
        read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
        uudecode(line, &buf[readcount], &thisread);
        for (int i = 0; i < thisread; i++)
            checksum += buf[readcount+i]&0xff;

        readcount += thisread;
    }

    read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    int checksum_sent = atoi(line);

    if (checksum_sent != checksum) {
        printf("checksum mismatch\n");
        write_string(lpc->fd, "RESEND\r\n");
        goto receive;
    }
    
    write_string(lpc->fd, "OK\r\n");

    return 0;
}

//////////////////////////////////////////////////////////////////////////
int lpc_write(lpcprog_t *lpc, uint32_t addr, int32_t len, char *data)
{
    char line[MAXLEN];

    if (write_string(lpc->fd, "W %u %d\r\n", addr, len))
        return -1;

    read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (line[0]!='0') {
        printf("lpc_write: %s",line);
        return -1;
    }

    int writecount;

retry :

    writecount = 0;

    while (writecount < len)
    {
        char tmp[MAXLEN];
        int tmplen;
        int thiswrite = min(len-writecount, MAX_UUENCODE_SIZE);
        uuencode(&data[writecount], thiswrite, tmp, &tmplen);
        write_string(lpc->fd, "%s\r\n", tmp);
        writecount += thiswrite;
    }

    int checksum = 0;
    for (int i = 0; i < len; i++)
        checksum += data[i]&0xff;

    write_string(lpc->fd,"%d\r\n", checksum);

    // XXX flush the write buffer here

    int res = read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (strcmp(line, "OK")) {
        printf("lpc_write: %d %s\n", res, line);
        goto retry;
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////
int lpc_copy_ram_to_flash(lpcprog_t *lpc, int32_t ram_addr, int32_t flash_addr, int32_t len)
{
    char line[MAXLEN];

    if (write_string(lpc->fd, "C %d %d %d\r\n", flash_addr, ram_addr, len))
        return -1;

    read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (line[0]!='0') {
        printf("timeout %s",line);
        return -1;
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////
int lpc_prepare_sectors(lpcprog_t *lpc, int sector_low, int sector_high)
{
    char line[MAXLEN];

    if (write_string(lpc->fd, "P %d %d\r\n", sector_low, sector_high))
        return -1;

    read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (line[0]!='0') {
        return -1;
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////
int lpc_erase_sectors(lpcprog_t *lpc, int sector_low, int sector_high)
{
    char line[MAXLEN];

    if (write_string(lpc->fd, "E %d %d\r\n", sector_low, sector_high))
        return -1;

    read_line_timeout_min(lpc->fd, line, MAXLEN, TIMEOUT_MS);
    if (line[0]!='0') {
        return -1;
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////
// this doesn't work. we'll have to upload a program to do this.
int lpc_set_memmap(lpcprog_t *lpc, int value)
{
    char buf[4];

    buf[0] = value;
    buf[1] = 0;
    buf[2] = 0;
    buf[3] = 0;

    return lpc_write(lpc, 0xe01fc040, 4, buf);
}

//////////////////////////////////////////////////////////////////////////
void draw_progress(const char *label, int width, double percent, int datasize, int64_t utime_start)
{
    int64_t utime_now = utime();
    double throughput = datasize / ((utime_now - utime_start + 1)/1000000.0) / 1024.0;
    int prog = width * percent;

    printf("\r%10s  [",label);
    for (int i = 0; i < prog; i++)
        printf("=");
    for (int i = 0; i < width-prog; i++)
        printf(" ");
    printf("] %5.1f %%  ( %5.2f kB/s )", percent*100, throughput);

    fflush(NULL);
}

//////////////////////////////////////////////////////////////////////////
int get_sector(lpcprog_t *lpc, uint32_t addr)
{
    for (int i = 0; i < lpc->device->numsectors; i++)
    {
        uint32_t lo = lpc->device->sectors[i].addr;
        uint32_t hi = lo + lpc->device->sectors[i].size;

        if (addr >= lo && addr < hi)
            return lpc->device->sectors[i].id;
    }

    //    warn("Couldn't find sector number for address!");
    return -1;
}

//////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[])
{
    lpcprog_t *lpc = (lpcprog_t*) calloc(1, sizeof(lpcprog_t));

    lpc->gopt = getopt_create();
    getopt_add_bool(lpc->gopt,  'h', "help",   0,              "Show this");
    getopt_add_int(lpc->gopt,   'b', "baud",   "57600",        "Baud rate");
    getopt_add_string(lpc->gopt,'d', "device", "/dev/ttyUSB0", "Device to connect to");
    getopt_add_bool(lpc->gopt,    0, "raw",    0,              "Treat the device as a raw device, not a tty");
    getopt_add_string(lpc->gopt,  0, "part",    "auto",        "The target device, or 'list' to list all known devices.");
    getopt_add_int(lpc->gopt,     0, "khz",    "12000",        "Oscillator speed of target board (in kHz)");

    getopt_add_bool(lpc->gopt,   0, "write-fixup",   1,       "Fix up the vector table checksum");
    getopt_add_bool(lpc->gopt,  'w', "write",   0,             "Write to flash");
    getopt_add_bool(lpc->gopt,  'v', "verify",  0,             "Verify data");
    getopt_add_bool(lpc->gopt,  'r', "reboot",  0,             "Reboot after upload");
    getopt_add_bool(lpc->gopt,    0, "verify-silence-bootblock", 1, "Ignore verification errors in first 64 bytes of bootblock");
    getopt_add_bool(lpc->gopt,  't', "terminal", 0,            "Stay in terminal mode");
    getopt_add_bool(lpc->gopt,    0, "terminal-echo", 0,       "Echo STDIN in terminal mode");
    getopt_add_int(lpc->gopt,     0, "terminal-baud", "115200", "Baud rate for terminal mode");

    // We support writing 252 bytes of random data to a designated
    // portion of flash.  The data is preceeded by 4 bytes of
    // signature data. If the signature is already present, the existing
    // random data will not be overwritten.
    getopt_add_bool(lpc->gopt, 'i', "write-id", 0,              "Write a unique ID to the flash");
    getopt_add_string(lpc->gopt, 0, "write-id-signature", "0x38598ad0",  "Signature to preface ID");
    getopt_add_string(lpc->gopt, 0, "write-id-address", "0x007c000",  "Flash address of signature");
    
    if (!getopt_parse(lpc->gopt, argc, argv, 1) || getopt_get_bool(lpc->gopt,"help")) {
        printf("Usage: %s [-wvrt] [options] <file.srec>\n\n", argv[0]);
        getopt_do_usage(lpc->gopt);
        return 0;
    }

    if (!strcmp(getopt_get_string(lpc->gopt, "part"), "list")) {
        list_devices(lpc);
        return 0;
    }

    /* The program has the following steps:

    1. Read the image file.
    2. Connect to LPC
    3. Flash the image (optional)
    4. Verify the image (optional)
    5. Write unique IDs, if unique ID is not found (optional)
    5. Reboot the LPC (optional)
    6. Provide a basic terminal mode to the serial port (optional)
    */

    ////////////// connect to target board ////////////////
    if (getopt_get_bool(lpc->gopt, "raw"))
        lpc->fd = open(getopt_get_string(lpc->gopt, "device"), O_RDWR | O_NOCTTY);
    else
        lpc->fd = serial_open(getopt_get_string(lpc->gopt, "device"),
                              getopt_get_int(lpc->gopt, "baud"),
                              1);
    
    if (lpc->fd < 0) 
    {
        perror(getopt_get_string(lpc->gopt, "device"));
        return 0;
    }

    objdata_t *obj;
    if (getopt_get_bool(lpc->gopt,"write") || getopt_get_bool(lpc->gopt,"verify"))
    {
        ////////////// read the image file ////////////////
        /* each parser should produce a list of chunks. The chunks should be:
           1. In increasing order of address
           2. A full CHUNKSIZE bytes (zero padded, if need-be)
           3. At an address on a multiple of CHUNKSIZE bytes
        */
        char *imagepath = vector_get(lpc->gopt->extraargs, 0);
        if (imagepath == NULL)
        {
            printf("No image file specified, exiting.\n");
            return -1;
        }

        obj = NULL;
    
        // SREC file?
        if (strstr(imagepath, ".srec"))
            obj = srec_parse(vector_get(lpc->gopt->extraargs, 0), CHUNKSIZE);
    
        // iHEX file? (not implemented)
        //  if (strstr(imagepath, ".hex"))
        //     obj = hex_parse(vector_get(lpc->gopt->extraargs, 0), CHUNKSIZE); 
    
        if (obj == NULL)
            fail("Unable to load image file (unrecognized suffix. Should be .srec)");
    
        printf("Image file loaded: 0x%08x bytes, checksum: 0x%08x\n", obj->nchunks*CHUNKSIZE, chunks_checksum(obj->chunks));

        if (lpc_sync(lpc))
            fail("Couldn't synchronize to LPC device");
        
        if (lpc_unlock(lpc))
            fail("Couldn't unlock LPC flash");
    } else {
      serial_set_dtr(lpc->fd, 0);
      serial_set_rts(lpc->fd, 0);
    }

    ////////////////// flash //////////////////
    if (getopt_get_bool(lpc->gopt, "write"))
    {
        // erase
        int64_t utime_start = utime();
        int lastsector = -1;
        int chunkserased = 0;
        for (chunk_t *c = obj->chunks; c != NULL; c = c->next)
        {
            int sector = get_sector(lpc, c->addr);
            if (sector != lastsector && sector >= 0)
            {
//                printf("erase sector %i\n", sector);
                lpc_prepare_sectors(lpc, sector, sector);
                lpc_erase_sectors(lpc, sector, sector);
                lastsector = sector;
            }

            chunkserased++;
            draw_progress("Erasing", PROGRESS_WIDTH, chunkserased / ((double) obj->nchunks), chunkserased*CHUNKSIZE, utime_start);
        }
        printf("\n");

        // now write the flash
        utime_start = utime();
        int chunkswritten = 0;
        for (chunk_t *c = obj->chunks;  c != NULL; c = c->next)
        {
            int sector = get_sector(lpc, c->addr);

            if (sector < 0)
                continue;

            // fixup bootloader checksum
            if (c->addr == 0 && getopt_get_bool(lpc->gopt, "write-fixup"))
            {
                int32_t *p = (int32_t*) c->data;
                int32_t v = 0;
                
                for (int i = 0; i < 8; i++)
                    v += p[i];
                
                p[5] += (~v)+1;
            }
            
            /*
            printf("\n%08x : ", c->addr);
            for (int i = 0; i < c->datalen; i++) {
                if (i%16==0 && i > 0)
                    printf("           ");
                printf("%02x ", c->data[i] & 0xff);
                if (i%16==15 || i == c->datalen-1) {
                    printf("\n");
                }
            }
            */

            if (lpc_write(lpc, lpc->device->ramaddr, c->datalen, c->data))
                printf("error\n");
            
            if (lpc_prepare_sectors(lpc, sector, sector))
                printf("couldn't prepare sectors\n");
            
            if (lpc_copy_ram_to_flash(lpc, lpc->device->ramaddr, c->addr, c->datalen))
                printf("flash fail");
            
            chunkswritten++;
            draw_progress("Writing", PROGRESS_WIDTH, chunkswritten / ((double) obj->nchunks), chunkswritten*CHUNKSIZE, utime_start);
        }

        printf("\n");
    }

    ////////////////// verify //////////////////
    if (getopt_get_bool(lpc->gopt, "verify"))
    {
        int64_t utime_start = utime();
        int chunksread = 0;

        for (chunk_t *c = obj->chunks;  c != NULL; c = c->next)
        {
            if (get_sector(lpc, c->addr) < 0)
                continue;

            char chkbuf[MAXLEN];
            if (lpc_read(lpc, c->addr, c->datalen, chkbuf))
                printf("verify read fail\n");

            for (int i = 0; i < c->datalen; i++)
                if (chkbuf[i]!=c->data[i]) {
                    if (c->addr==0 && i < 64 && getopt_get_bool(lpc->gopt, "verify-silence-bootblock"))
                        continue;

                    printf("verify failed at %08x : wrote %02x, read %02x\n", 
                           c->addr+i, c->data[i]&0xff, chkbuf[i]&0xff); 
                }

            chunksread++;
            draw_progress("Verifying", PROGRESS_WIDTH, chunksread / ((double) obj->nchunks), chunksread*CHUNKSIZE, utime_start);
        }

        printf("\n");
    }

    ////////////////// write unique id ////////////////
    if (getopt_get_bool(lpc->gopt, "write-id")) {
        uint32_t addr = strtod(getopt_get_string(lpc->gopt, "write-id-address"), NULL);
        uint32_t signature = strtod(getopt_get_string(lpc->gopt, "write-id-signature"), NULL);

        uint8_t buf[256];
        if (lpc_read(lpc, addr, 256, (char*) buf)) 
            printf("Couldn't read old id\n");

        uint32_t current_signature = ntohl(((uint32_t*) buf)[0]);
        if (current_signature == signature) {
            printf("Keeping ID at %08x: %02x%02x%02x%02x...\n", addr, buf[5], buf[6], buf[7], buf[8]);
        } else {
            ((uint32_t*) buf)[0] = htonl(signature);
            struct timeval tv;
            gettimeofday(&tv, NULL);
            srand(tv.tv_usec ^ tv.tv_sec);

            for (int i = 4; i < 256; i++)
                buf[i] = rand();

            int sector = get_sector(lpc, addr);
            lpc_prepare_sectors(lpc, sector, sector);
            lpc_erase_sectors(lpc, sector, sector);
            
            lpc_write(lpc, lpc->device->ramaddr, 256, (char*) buf);
            lpc_prepare_sectors(lpc, sector, sector);
            if (lpc_copy_ram_to_flash(lpc, lpc->device->ramaddr, addr, 256))
                printf("Couldn't write new signature bytes\n");
            else
                printf("Wrote new signature ID bytes at address %08x: %02x%02x%02x%02x...\n", 
                       addr, buf[5], buf[6], buf[7], buf[8]);
        }
    }

    ////////////////// reboot //////////////////
    if (getopt_get_bool(lpc->gopt, "reboot"))
    {
        int64_t utime_start = utime();
        draw_progress("Rebooting", PROGRESS_WIDTH, 1, 0, utime_start);

        serial_set_dtr(lpc->fd, 1); // reset the board, not asserting p2.10 low
        serial_set_rts(lpc->fd, 0);
        usleep(200000);
        serial_set_dtr(lpc->fd, 0); // release reset

        printf("\n");
    }

    ////////////////// terminal //////////////////
    if (getopt_get_bool(lpc->gopt, "terminal"))
    {
        struct termios term;
        tcgetattr(STDIN, &term);
        term.c_lflag &= ~ICANON;
        if (!getopt_get_bool(lpc->gopt, "terminal-echo"))
            term.c_lflag &= ~ ECHO;
        tcsetattr(STDIN, TCSANOW, &term);
        setbuf(stdin, NULL);

        struct pollfd pfds[2];

        pfds[0].fd = STDIN;
        pfds[0].events = POLLIN;
        pfds[1].fd = lpc->fd;
        pfds[1].events = POLLIN;

        printf("----------------------terminal-----------------------\n");

        if (!getopt_get_bool(lpc->gopt, "raw"))
            serial_setbaud(lpc->fd, getopt_get_int(lpc->gopt, "terminal-baud"));

        while (1)
        {
            poll(pfds, 2, -1);
            char buf[MAXLEN];

            for (int i = 0; i <= 1; i++)
            {
                int j = 1-i;

                if (pfds[i].revents&POLLIN) {
                    ssize_t len = read(pfds[i].fd, buf, MAXLEN);
                    write(pfds[j].fd, buf, len);
                }
            }
        }
    }
}
